from confluent_kafka import Producer, Consumer, TopicPartition, KafkaError
from loguru import logger
from  ..lib.process import KillError
import time
import json
import random

def check_builtin_condition(key, now, end):
    if key == "depth":
        if now > end:
            return False
        else:
            return True
    elif key == "create_at":
        if time.time() - int(now) > int(end):
            return False
        else:
            return True




class KafkaSender:
    def __init__(self, servers: str, flush_timeout=10):
        """
        Init a kafka sender
        :param servers: (str)bootstrap.servers
        :param flush_timeout: (int)Timeout for flush message
        """
        self.servers = servers
        self.message_list = []
        self.flush_timeout = flush_timeout

    def add_message(self, topic: str, message, partition=None):
        """
        Add a message to the waiting queue
        :param topic: (str) the topic of the message
        :param message: (dict, list, str, bytes) the body of message
        :param partition: (int) the partition of the sending topic, default means random
        """
        if type(message) == dict or type(message) == list:
            msg = json.dumps(message).encode()
        elif type(message) == str:
            msg = message.encode()
        elif type(message) == bytes:
            msg = message
        else:
            raise TypeError("Unsupported type :" + type(message))
        self.message_list.append((topic, partition, msg))
        return

    def send_all(self) -> (bool, str):
        """
        Send all messages in waiting queue
        :return: (status, error_message)
        """
        if len(self.message_list) == 0:
            return True, "No available message"
        producer = Producer({'bootstrap.servers': self.servers})
        for message in self.message_list:
            producer.produce(topic=message[0], value=message[2], partition=message[1])
            producer.poll(0)
        producer.flush(self.flush_timeout)
        # producer.length == 0 means that all the messages have been sent.
        if len(producer) == 0:
            self.message_list = []
            return True, ""
        else:
            l1 = len(self.message_list)
            l2 = len(producer)
            self.message_list = []
            return False, "Not all messages has been sent. Sent(%d/%d)" % (l1 - l2, l1)


class KafkaConsumer(Consumer):
    def __init__(self, servers: str, topics: list, group: str):
        """
        Init a kafka consumer
        :param servers: (str)bootstrap.servers
        :param topics: (str)group.id
        :param group: (list of str)subscribe topics
        """
        super().__init__({
            'bootstrap.servers': servers,
            'group.id': 'autumn' + str(random.randint(0,9999999999)),
            'auto.offset.reset': 'earliest',
            'enable.auto.commit': False,
            'max.poll.interval.ms': 86400000
        })

    def __del__(self):
        self.close()

    def clear_assignment(self):
        """
        Clear all assignment
        """
        self.assign([])

    def add_assignment(self, topic, partition, offset):
        """
        Add a assignment
        :param topic: (str) Assignment topic
        :param partition: (int) Assignment partition
        :param offset: (int) Assignment offset
        """
        p = TopicPartition(topic=topic, partition=partition, offset=offset)
        self.assign(self.assignment() + [p])

    def remove_assignment(self, topic, partition, offset):
        """
        remove a assignment
        :param topic: (str) Assignment topic
        :param partition: (int) Assignment partition
        :param offset: (int) Assignment offset
        """
        self.assign(self.assignment().remove(TopicPartition(topic=topic, partition=partition, offset=offset)))

    def poll(self):
        """
        Poll a message
        :return: (confluent_kafka.Message) Kafka message
        """
        while True:
            message = super().poll(5)
            if message is None:
                continue
            if message.error():
                if message.error().code() == -191:
                    time.sleep(1)
                    continue
                raise KafkaError(message.error())
            return message


class KafkaMonitor(KafkaConsumer):
    def __init__(self, executor):
        """
        Init a kafka monitor
        :param kafka_url: kafka urls
        :param topic: kafka_topic
        :param group: kafka_group
        :param arrangement: the partition&offset arrangement to sync
        """
        self.executor=executor
        logger.debug(id(self.executor))
        super().__init__(executor.kafka_url, [executor.kafka_topic], executor.kafka_group)
        self.clear_assignment()
        self.topic = executor.kafka_topic

    def poll(self):
        """
        Get a new message
        :return: confluent_kafka.Message
        """
        while True:
            # 出现了新的分配/再平衡
            temp = None
            while True:
                try:
                    temp = self.executor.queue.get_nowait()
                except:
                    break
            if temp:
                if temp == "kill":
                    raise KillError("Killed")
                self.executor.kafka_arrangement = temp
                if self.executor.kafka_arrangement["partition"] != "":
                    self.assign([TopicPartition(self.topic, self.executor.kafka_arrangement["partition"], self.executor.kafka_arrangement["offset_range"][0])])

            if not self.executor.kafka_arrangement or not self.executor.kafka_arrangement or self.executor.kafka_arrangement["partition"] == "":
                time.sleep(5)
                continue
            if self.executor.kafka_arrangement["offset_range"][0] <= self.executor.kafka_arrangement["offset_range"][1]:
                self.assign([TopicPartition(self.topic, self.executor.kafka_arrangement["partition"], self.executor.kafka_arrangement["offset_range"][0])])
            else:
                self.executor.kafka_arrangement = {}
                continue
            self.executor.kafka_arrangement["offset_range"][0] += 1

            return super().poll()



from tenacity import retry, wait_fixed, stop_after_attempt
import requests
from loguru import logger
import os
import time



def event_process(trigger, task, params = None, callback = None):
    params = {} if params is None else params
    trigger.bind_task(task)
    trigger.emit(**params)
    if callback:
        callback(os.getpid())

def custom_process(trigger, task, params = None, callback = None):
    params = {} if params is None else params
    trigger.bind_task(task)
    trigger.run(**params)
    if callback:
        callback(os.getpid())

def time_process(trigger, interval, task, params = None):
    params = {} if params is None else params
    trigger.bind_task(task)
    while True:
        trigger.emit(**params)
        time.sleep(interval)



class Trigger:
    DEFAULT_INTERVAL = 10

    def __init__(self, name: str, trigger_type: str, interval=None, event_name=None):
        """
        Init a trigger
        :param name: (str) Trigger's name, not repeatable
        :param trigger_type: The type of trigger, must be one of ["time","event","custom"]
        :param interval: Time interval for time trigger. Invalid for other types of triggers
        :param event_name: Event name for event trigger. Invalid for other types of triggers
        """
        # check params
        self.name = name
        if trigger_type not in ["time", "event", "custom"]:
            raise TypeError("Unsupported trigger type " + trigger_type)
        self.type = trigger_type
        if trigger_type == "time":
            if interval is None:
                print("Warning: Time trigger %s is short of interval, use default %d seconds" % (
                    name, self.DEFAULT_INTERVAL))
                self.interval = self.DEFAULT_INTERVAL
            else:
                if interval <= 0:
                    raise ValueError("Invalid interval %d, It must >=0" % interval)
                else:
                    self.interval = interval
        elif trigger_type == "event":
            if event_name is None:
                raise ValueError("Invalid event name")
            else:
                self.event = str(event_name)

        # init properties
        self.function = self.default_function
        # init from app.dump
        self.controller_url = None
        self.executor = None

    def default_function(self, *args, **kwargs):
        if not args and not kwargs:
            raise ValueError("Empty input")
        elif args:
            return args[0]
        else:
            return kwargs[kwargs.keys()[0]]



    def run(self):
        """
        Only valid when type is custom, rewrite by user, run as a process
        """
        pass


    def bind(self, func):
        """
        Bind a function for this trigger to emit
        :param func: A function or other callable object
        """
        if not hasattr(func, "__call__"):
            raise TypeError("You can only bind a callable object")
        self.function = func

    def bind_task(self, task):
        self.__task = task

    @retry(stop=stop_after_attempt(3), wait=wait_fixed(2))
    def report(self, data):
        """
        Report self status to controller
        :param data:
        :return:
        """


    def emit(self, *args, **kwargs):
        """
        Emit this trigger, and generate executor inputs
        :param args: params for bound function
        :param kwargs: params for bound function
        """
        if self.function is None:
            raise ValueError("Unbind function")
        try:
            data = self.function(*args, **kwargs)
        except:
            logger.error("Error in emit binded function: \n", exc_info=True)
            return
        try:
            self.executor.data_process(self.__task, data)
        except:
            logger.error("Error in handle data: \n", exc_info=True)
